/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of rbtree
 *
 * rbtree is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rbtree is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/*
  a complete implementation of balanced binary trees using the red-black
  algorithm.  Algorithms were taken from "Introduction to Algorithms",
  by Thomas H. Cormen, Charles E. Leiserson, and Ronald L. Rivest.
*/

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>

#include "rbtreep.h"

/* macros to easily access node and tree members */
#define LEFT_SIB(node)          ((node)->descendent[LEFT])
#define RIGHT_SIB(node)         ((node)->descendent[RIGHT])
#define DESCEND(node,dir)	((node)->descendent[dir])
#define PARENT(node)            ((node)->parent)
#define COLOR(node)             ((node)->p.color)
#define DATA(node)              ((node)->data)
#define ROOT(tree)              ((tree)->root)
#define COUNT(tree)             ((tree)->count)
#define NIL(tree)		(&(tree)->nil)
#define CMP(tree)		((tree)->cmp)

/*

  PROCEDURE
  rbtree_new - Create the root of a red-black tree.

  DESCRIPTION
  This routine creates the data structures necessary to implement the
  root of a red-black tree.  It requires a node comparison function
  which will be used to sort the nodes.

  The comparison routine is usually only used for sorting, not
  searching, but may be if a search or destroy routine is called
  without a specific comparison routine.  It should return @samp{-1},
  @samp{0}, or @samp{1} depending upon whether the first node is
  respectively less than, equal to, or greater than the second node.

  RETURNS
  It returns a handle to the new tree if successful, @code{NULL} if not.

*/
RBTree
rbtree_new(
  int (*cmp)(const void *, const void *)
	/* a node comparison routine used to sort the nodes */
)
{
  Tree *tree = (Tree *) malloc(sizeof(Tree));

  if (tree)
  {
    tree->cmp = cmp;
    ROOT(tree) = NIL(tree);
    COUNT(tree) = 0;
    LEFT_SIB(NIL(tree)) = RIGHT_SIB(NIL(tree)) = NIL(tree);
    COLOR(NIL(tree)) = BLACK;
  }

  return (RBTree) tree;
}

/*

  PROCEDURE
  rbtree_delete - Delete a red-black binary tree

  DESCRIPTION
  This routine deletes a red-black binary tree, destroying all nodes
  in the specified order and direction.  It optionally takes a user
  supplied function which, when passed a node's data pointer,
  deallocates user node data.  It is strongly suggested that
  the deallocation function be specified.

*/
void
rbtree_delete(
  RBTree rbtree,		/* the handle of the rbtree to delete */
  void (*nfree)(void *),	/* a routine which deallocates node user
				   memory, may be @code{RBTREE_NULL_DELETE} */
  Visit visit,		/* the order of deletion, either
			   @code{IN_ORDER}, or @code{POST_ORDER},
			   (@code{PRE_ORDER} is treated as
			   @code{POST_ORDER}) */
  SiblingOrder sibling_order	/* the direction that the tree is parsed */
)
{
  Tree *tree = (Tree *) rbtree;

  if ( IN_ORDER == visit )
    Free_in_order(tree, ROOT(tree), nfree, sibling_order);
  else
    Free_post_order(tree, nfree, sibling_order);

  free( rbtree );
}

/*

  PROCEDURE
  rbtree_udelete - Delete a red-black binary tree.

  DESCRIPTION
  This routine deletes a red-black binary tree, destroying all nodes
  in the specified order and direction.  It optionally takes a user
  supplied function which, when passed a node's data pointer,
  deallocates possible user node data.  It is strongly suggested that
  the deallocation function be specified.  This routine differs from
  @code{rbtree_delete} in that it can pass along a pointer provided by
  the calling routine to the deallocation function routine, allowing
  arbitrary data to be available to it.

*/
void
rbtree_udelete(
  RBTree rbtree,		/* the handle of the rbtree to delete */
  void (*nfree)(void *, void *),/* a routine which deallocates node user
				   memory, may be @code{RBTREE_NULL_DELETE} */
  void *udata,			/* a pointer to data to be passed to the
				   action routine */
  Visit visit,			/* the order of deletion.  (@code{PRE_ORDER} is
				   treated as @code{POST_ORDER}) */
  SiblingOrder sibling_order	/* the direction that the tree is parsed */
)
{
  Tree *tree = (Tree *) rbtree;

  if ( IN_ORDER == visit )
    uFree_in_order(tree, ROOT(tree), nfree, udata, sibling_order);
  else
    uFree_post_order(tree, nfree, udata, sibling_order);

  free( rbtree );
}

/*

  PROCEDURE
  rbtree_walk - Walk along a tree, processing each node.

  DESCRIPTION
  This routine traverses an @code{rbtree}, calling a user supplied action
  function at selectable traversals of each node (any combination of
  pre-order, in-order, or post-order).  The function is informed of
  which node traversal and which tree level it is called from.  The
  walk is aborted if the action routine returns non-zero.

  The traversals at which the action routine is called at are
  specified by the logical or of @code{PRE_ORDER}, @code{IN_ORDER}, or
  @code{POST_ORDER}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
rbtree_walk(
  RBTree rbtree,	/* a handle to the rbtree to traverse */

  /* the user supplied action routine applied to each node */
  int (*action)(void * nd, Visit visit, unsigned long level),

  SiblingOrder sibling_order,	/* The direction in which the tree is
				   traversed. */
  Visit visit			/* The node traversals at which to call the
  				   action routine. The logical or of
  				   possible @code{Visit} values. */
)
{
  Tree *tree = (Tree *) rbtree;

  return Walk(tree, ROOT(tree), action, sibling_order, visit, 0);
}

/*

  PROCEDURE
  rbtree_uwalk - Walk along a tree, processing each node.

  DESCRIPTION
  This routine traverses an @code{rbtree}, calling a user supplied
  action function at selectable traversals of each node (any
  combination of pre-order, in-order, or post-order).  The function is
  informed of which node traversal and which tree level it is called
  from.  The walk is aborted if the action routine returns non-zero.
  This routine differs from @code{rbtree_walk} in that it can pass
  along a pointer provided by the calling routine to the action
  routine, allowing arbitrary data to be available to the action
  routine.

  The traversals at which the action routine is called at are
  specified by the logical or of @code{PRE_ORDER}, @code{IN_ORDER},
  or @code{POST_ORDER}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
rbtree_uwalk(
  RBTree rbtree,		/* a handle to the rbtree to traverse */

  /* the user supplied action routine applied to each node */
  int (*action)(void * nd, Visit visit, unsigned long level, void *udata),

  void *udata,			/* a pointer to data to be passed to the action
				   routine */

  SiblingOrder sibling_order,	/* the direction in which the tree is
				   traversed */
  Visit visit			/* The node traversals at which to call the
  				   action routine. The logical or of
  				   possible @code{Visit} values. */
)
{
  Tree *tree = (Tree *) rbtree;

  return uWalk(tree, ROOT(tree), action, udata, sibling_order, visit, 0);
}

/*

  PROCEDURE
  rbtree_traverse - Traverse a red-black binary tree in order, processing
                      each node.

  DESCRIPTION
  This routine walks along an @code{rbtree}, calling a user supplied
  action function at the in-order traversals of each node.  The tree
  traversal is aborted if the action routine returns non-zero.
  @code{rbtree_traverse} is substantially faster than calling
  @code{rbtree_walk} with an in-order visit, and uses fewer resources.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
rbtree_traverse(
  RBTree rbtree,		/* a handle to the rbtree to traverse */
  int (*action)(void * nd),	/* the user supplied action routine applied
  				   to each node */
  SiblingOrder sibling_order	/* the direction in which the tree is
				   traversed */
)
{
  Tree *tree = (Tree *) rbtree;

  return Traverse(tree, ROOT(tree), action, sibling_order);
}


/*

  PROCEDURE
  rbtree_utraverse - Traverse a red-black binary tree in order, processing
                      each node.

  DESCRIPTION
  This routine walks along an @code{rbtree}, calling a user supplied
  action function at the in-order traversals of each node.  The tree
  traversal is aborted if the action routine returns non-zero.
  @code{rbtree_utraverse} is substantially faster than calling
  @code{rbtree_uwalk} with an in-order visit, and uses fewer
  resources.  This routine differs from @code{rbtree_traverse} in that
  it can pass along a pointer provided by the calling routine to the
  action routine, allowing arbitrary data to be available to the
  action routine.
    
  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
rbtree_utraverse(
  RBTree rbtree,		/* a handle to the rbtree to traverse */
  int (*action)(void *nd, void *udata ),	/* the user supplied action
						   routine applied to each
						   node */
  void *udata,			/* a pointer to data to be passed to the action
				   routine */
  SiblingOrder sibling_order	/* the direction in which the tree is
				   traversed */
)
{
  Tree *tree = (Tree *) rbtree;

  return uTraverse(tree, ROOT(tree), action, udata, sibling_order);
}


/*

  PROCEDURE
  rbtree_count - Determine the number of nodes in a red-black tree.

  DESCRIPTION
  This routine returns the number of nodes in an @code{rbtree}.
  As each tree keeps a counter of how many nodes is in it, this
  is an inexpensive function to call.

*/
unsigned long
rbtree_count(
  RBTree rbtree		/* the tree in question */
)
{
  return COUNT((Tree *)rbtree);
}


/*

  PROCEDURE
  rbtree_insert - Create and insert a node into a red-black tree.

  DESCRIPTION
  @code{rbtree_insert} creates a node, sets its data pointer to the
  specified pointer, and inserts it into the specified tree.

  RETURNS
  It returns @samp{0} if the insert was successful, @samp{1} if it was
  unable to create the new node.

*/
int
rbtree_insert(
  RBTree rbtree,	/* a handle to the tree into which to insert the
			   node */
  void *data		/* a pointer to the data */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = newNode(tree, data);
  if (! node) return 1;

  Insert(tree, node);
  return 0;
}


/*

  PROCEDURE
  rbtree_insert_dnode - Insert a detached node into a tree.

  DESCRIPTION
  This routine inserts a detached node into a binary tree. It assumes
  that the tree's insert/delete comparison function can be applied to
  the data in the passed node.  The node handle must have been
  obtained from @code{rbtree_detach_node}.

  RETURNS
  It returns @samp{1} if the node is @code{NULL} or @code{NIL(tree)},
  @samp{0} otherwise.

*/
int
rbtree_insert_dnode(
  RBTree rbtree,	/* a handle to the tree into which to insert
			   the node */
  RBNode rbnode		/* the node to insert */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = (Node *) rbnode;
  if (! node || node == NIL(tree)) return 1;

  Insert(tree, node);
  return 0;
}

/*

  PROCEDURE
  rbtree_destroy - Search for a node in a red-black binary tree and
  destroy it.

  DESCRIPTION
  @code{rbtree_destroy} searches a binary tree for the node for which
  the passed data compares equivalently with the node's data.  It uses
  the passed comparison routine, if available.  If not, it uses that
  with which the tree was initialized.  In the former case the passed
  data need not have the same form as the data stored in the node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  @code{rbtree_destroy} frees the memory required by the node.  The
  user must destroy the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer if the node was found,
  @code{NULL} otherwise.

*/
void *
rbtree_destroy(
  RBTree rbtree,	/* the red-black tree to search */
  const void *data,	/* the data to search for */
  int (*cmp)(const void *, const void *)
  			/* The address of a comparison function.  Set to
			   @code{RBTREE_NULL_CMP} to use the tree's initial
			   comparison function. */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = Search(tree, ROOT(tree), data, cmp ? cmp : CMP(tree));

  if (node != NIL(tree))
  {
    void *datap = DATA(node);
    deleteNode(Delete(tree, node));
    return datap;
  }

  return NULL;
}


/*

  PROCEDURE
    rbtree_node_size - Return the size of an @code{RBNode}.

  DESCRIPTION
    This routine returns the size of the internal structure used for
    each node.  This does not include user supplied data.  Please
    note that this is a function call, not a macro!

*/
size_t
rbtree_node_size(
  void
)
{
  return sizeof(Node);
}

/*

  PROCEDURE
  rbtree_destroy_node - Remove a node from a binary tree and destroy it.

  DESCRIPTION
  This routine removes the passed node from the specified tree and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  Returns the node's data pointer if the node is not @code{NULL} or
  @code{NIL(tree)}, @code{NULL} otherwise.

*/
void *
rbtree_destroy_node(
  RBTree rbtree,	/* the tree from which to remove the node */
  RBNode rbnode		/* the node to remove */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = (Node *) rbnode;

  if (node && node != NIL(tree))
  {
    void *datap = DATA(node);
    deleteNode(Delete(tree, node));
    return datap;
  }

  return NULL;
}


/*

  PROCEDURE
  rbtree_detach_node - Remove a node from an red-black tree without
  destroying it.

  DESCRIPTION
  @code{rbtree_detach_node} detaches the specified node from the specified
  tree, but doesn't delete it.  This is useful for inserting the node into
  another tree.  Because of the way nodes are deleted from trees, the node
  actually removed may not be the node passed to the routine.  The user
  supplied data is correctly tracked.

  RETURNS
  Returns the node that was actually detached, or @code{NULL} if the
  node is @code{NULL} or @code{NIL(tree)}.

*/
RBNode
rbtree_detach_node(
  RBTree rbtree,	/* the tree from which to remove the node */
  RBNode rbnode		/* the node to detach */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = (Node *) rbnode;

  if (node && node != NIL(tree))
  {
    void *nd = DATA(node);

    node = Delete(tree, node);
    DATA(node) = nd;
    return (RBNode) node;
  }

  return NULL;
}



/*

  PROCEDURE
  rbtree_destroy_dnode - Deallocate the memory associated with a
  detached node.

  DESCRIPTION
  This routine deallocates the memory associated with a detached
  red-black tree node.  The user must have already destroyed the data
  associated with the node.  The node handle must have been returned
  by @code{rbtree_detach_node}.

*/
void
rbtree_destroy_dnode(
  RBNode rbnode		/* the handle of the detached node to destroy */
)
{
  Node *node = (Node *) rbnode;

  if (node) deleteNode(node);
}



/*

  PROCEDURE
  rbtree_search - Search a red-black tree for a node with equivalent data.

  DESCRIPTION
  @code{rbtree_search} searches the tree for the node for which the
  passed data compares equivalently with the node's data.  It uses the
  passed comparison routine, if available.  If not, it uses that with
  which the tree was initialized.  In the former case the passed data
  need not have the same form as the data stored in the node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns the node's data pointer if the node was found,
  @code{NULL}  otherwise.

*/
void *
rbtree_search(
  RBTree rbtree,	/* the red-black tree to search */
  const void *data,		/* the data to search for */
  int (*cmp)(const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{RBTREE_NULL_CMP} to use the tree's initial
		 comparison function. */
)
{
  Tree *tree;
  Node *node;

  tree = (Tree *) rbtree;
  node = Search(tree, ROOT(tree), data, cmp ? cmp : CMP(tree));

  return (node != NIL(tree)) ? DATA(node) : NULL;
}


/*

  PROCEDURE
  rbtree_search_node - Search a red-black tree for a node with
  equivalent data.

  DESCRIPTION
  This routine searches a red-black tree for the node for which the
  passed data compares equivalently with the node's data.  It uses the
  passed comparison routine, if available.  If not, it uses that with
  which the tree was initialized.  Note that in the former case the
  passed data need not have the same form as the data stored in the
  node.  The comparison routine is called with the passed node's data
  as the first argument and the tree node's data as the second
  argument.  It must return @samp{-1}, @samp{0}, or @samp{1} if,
  respectively, the first argument is less than, equal to, or greater
  than the second.

  RETURNS
  It returns a handle to the matching node, or @code{NULL} if not found.

*/
RBNode
rbtree_search_node(
  RBTree rbtree,		/* the tree to search */
  const void *data,		/* the data to match */
  int (*cmp)(const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{RBTREE_NULL_CMP} to use the tree's initial
		 comparison function. */
)
{
  Tree *tree;
  Node *node;

  tree = (Tree *) rbtree;
  node = Search(tree, ROOT(tree), data, cmp ? cmp : CMP(tree));
  return (node != NIL(tree)) ? (RBNode) node : NULL;
}


/*

  PROCEDURE
  rbtree_bnd_search - Search a red-black binary tree keeping track of
  sibling nodes.

  DESCRIPTION
  This routine searches a binary tree for the node for which the
  passed data compares equivalently with the node's data.  It uses the
  passed comparison routine, if available.  If not, it uses that with
  which the tree was initialized.  Note that in the former case the
  passed data need not have the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  As the tree is being searched, @code{rbtree_bnd_search} keeps track
  of the preceding and succeeding in-order nodes.  If the node is not
  found, the nodes which bracket the data are returned.  The result of
  the search is returned via the parameters @code{prev}, and
  @code{next}.  If a node has data which matches the key data, both
  parameters are set to the node's data pointer.  If a node is not
  found, and if there exists a previous in-order node, the parameter
  @code{prev} is set to that node's data pointer, else it is set to
  @code{NULL}.  If there exists a succeeding in-order node, the
  parameter @code{next} is set to its data pointer, else it is set to
  @code{NULL}.

  RETURNS
  It returns the node's data pointer if the node can be found,
  @code{NULL} otherwise.

*/
void *
rbtree_bnd_search(
  RBTree rbtree,	/* the red-black tree to search */
  const void *data,	/* the data to search for */
  void **prev,		/* holding place for data that is found */
  void **next,		/* holding place for data that is found */
  int (*cmp)(const void *, const void *)
	      		/* The address of a comparison function.  Set to
			   @code{RBTREE_NULL_CMP} to use the tree's initial
			   comparison function. */
)
{
  Tree *tree;
  Node *node, *next_node, *prev_node;

  tree = (Tree *) rbtree;
  node = BndSearch(tree, ROOT(tree), &prev_node, &next_node, data,
				  cmp ? cmp : CMP(tree));
  if (node != NIL(tree))
  {
    if (prev) *prev = DATA(node);
    if (next) *next = DATA(node);
    return DATA(node);
  }

  *next = (next_node == NIL(tree)) ? NULL : DATA(next_node);

  *prev = (prev_node == NIL(tree)) ? NULL : DATA(prev_node);

  return NULL;
}


/*

  PROCEDURE
  rbtree_bnd_search_node - Search a red-black binary tree keeping
  track of sibling nodes.


  DESCRIPTION
  This routine searches a binary tree for the node for which the
  passed data compares equivalently with the node's data.  It uses the
  passed comparison routine, if available.  If not, it uses that with
  which the tree was initialized.  Note that in the former case the
  passed data need not have the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  As the tree is being searched, @code{rbtree_bnd_search_node} keeps
  track of the preceding in-order and succeeding nodes.  If the node
  is not found, the nodes which bracket the data are returned.  The
  result of the search is returned via the parameters @code{prev} and
  @code{next}.  If a node has data which matches the key data, both
  parameters are set to point at the node.  If a node is not found,
  and if there exists a previous in-order node, the parameter
  @code{prev} is set to point at the node, else it is set to
  @code{NULL}.  If there exists a succeeding in-order node, the
  parameter @code{next} is set to point to it, else it is set to
  @code{NULL}.

  RETURNS
  It returns a handle to the found node, or @code{NULL} if not found.

 */
RBNode
rbtree_bnd_search_node(
  RBTree rbtree,		/* binary tree to search */
  const void *data,		/* the data to search for */
  RBNode *prev,			/* set to the previous node */
  RBNode *next,			/* set to the next node */
  int (*cmp)(const void *, const void *)
				/* The address of a comparison function.
				   Set to @code{RBTREE_NULL_CMP} to use the
				   tree's initial comparison function. */
)
{
  Tree *tree;
  Node *node;

  tree = (Tree *) rbtree;
  node = BndSearch(tree, ROOT(tree), (Node **) prev, (Node **)next,
				  data, cmp ? cmp : CMP(tree));

  if (node != NIL(tree))
  {
    *prev = *next = node;
    return node;
  }

  if (*prev == NIL(tree)) *prev = NULL;
  if (*next == NIL(tree)) *next = NULL;

  return NULL;
}



/*

  PROCEDURE
  rbtree_min - Determine the node with the minimum data.

  DESCRIPTION
  This routine searches the tree for the node with the minimum
  data, as determined by the comparison routine with which the tree
  was created.

  RETURNS
  It returns the node's data pointer, @code{NULL} if the tree has no data.

*/
void *
rbtree_min(
 RBTree rbtree		/* the red-black tree to scan */
)
{
  Tree *tree = (Tree *) rbtree;

  return (ROOT(tree) == NIL(tree)) ? NULL : DATA(Minimum(tree, ROOT(tree)));
}


/*

  PROCEDURE
  rbtree_min_node - Determine the node with the minimum data.

  DESCRIPTION
  This routine searches the tree for the node with the minimum
  data, as determined by the comparison routine with which the tree
  was created.

  RETURN
  It returns a handle to the node if it exists, @code{NULL} if the
  tree is empty.

*/
RBNode
rbtree_min_node(
 RBTree rbtree		/* the tree to search */
)
{
  Tree *tree = (Tree *) rbtree;

  return (ROOT(tree) != NIL(tree)) ? (RBNode) Minimum(tree,
						       ROOT(tree)) : NULL;
}

/*

  PROCEDURE
  rbtree_max - Determine the node with the maximum data.

  DESCRIPTION
  This routine searches the tree for the node with the maximum
  data, as determined by the comparison routine with which the tree
  was created.

  RETURNS
  It returns the node's data pointer, @code{NULL} if the tree has no data.

*/
void *
rbtree_max(
 RBTree rbtree		/* the red-black tree to scan */
)
{
  Tree *tree = (Tree *) rbtree;

  return (ROOT(tree) == NIL(tree)) ? NULL : DATA(Maximum(tree, ROOT(tree)));
}


/*

  PROCEDURE
  rbtree_max_node - Determine the node with the maximum data.

  DESCRIPTION
  This routine searches the tree for the node with the maximum
  data, as determined by the comparison routine with which the tree
  was created.

  RETURN
  It returns a handle to the node if it exists, @code{NULL} if the
  tree is empty.

*/
RBNode
rbtree_max_node(
 RBTree rbtree		/* the tree to search */
)
{
  Tree *tree = (Tree *) rbtree;

  return (ROOT(tree) != NIL(tree)) ?
    (RBNode) Maximum(tree, ROOT(tree)) : NULL;
}


/*

  PROCEDURE
  rbtree_next_node - Determine the next in-order node to a given node.

  DESCRIPTION
  Determine the next in-order node (depending upon the user's desired
  traversal direction).

  RETURNS
  It returns a handle to the next node or @code{NULL} if there is none.

*/
RBNode
rbtree_next_node(
  RBTree rbtree,		/* the tree to traverse */
  RBNode rbnode,		/* the node to find the neighbor of */
  SiblingOrder sib_order	/* the direction of traversal */
)
{
  Tree *tree = (Tree *) rbtree;
  Node *node = (Node *) rbnode;

  if (node == NIL(tree) || node == NULL) return NULL;

  node = (sib_order == LEFT_TO_RIGHT) ?
           Successor(tree, node) : Predecessor(tree, node);
  return (node != NIL(tree)) ? (RBNode) node : NULL;
}


/*

  PROCEDURE
  rbtree_node_get_data - Retrieve the data from a specified @code{rbtree} node.

  RETURNS
  It returns the node's data pointer.

*/
void *
rbtree_node_get_data(
  RBNode rbnode		/* the node from which to extract the data */
)
{
  Node *node = (Node *) rbnode;

  return DATA(node);
}


/*

  PROCEDURE
  rbtree_node_put_data - Replace data pointer in a specified
  @code{rbtree} node.

  DESCRIPTION
  This routine replaces the data pointer in the specified
  @code{rbtree} node with the passed pointer.  It does not attempt to
  resort the tree, so the new data should not change the node's
  position in the tree (if it's not detached).

*/
void
rbtree_node_put_data(
  RBNode rbnode,	/* the node into which to copy the data */
  void *data		/* the data to copy */
)
{
  Node *node = (Node *) rbnode;

  DATA(node) = data;
}


/*

  PROCEDURE
  rbtree_join - Join two red-black trees.

  DESCRIPTION
  This routine joins two redblack trees, moving nodes from the second tree
  to the first.  It doesn't delete the second tree, just empties it.  This
  routine is more efficient than calling @code{rbtree_detach_node} and
  @code{rbtree_insert_dnode}.

*/
void
rbtree_join(
  RBTree tree1,		/* the tree to merge the second into */
  RBTree tree2		/* the tree to merge into the first tree */
)
{
  JoinTrees((Tree *)tree1, (Tree *)tree2);
}


/* Static routines */

/* redirect parent node.  need not check for legal *<to>*, as to will point
   to NIL(tree) */
#define RedirectPARENT(tree, to, from) \
         {\
           PARENT(to) = PARENT(from); \
           if (PARENT(to) == NIL(tree)) ROOT(tree) = (to); \
           else if (LEFT_SIB(PARENT(to)) == (from))  LEFT_SIB(PARENT(to)) = (to); \
           else                                 RIGHT_SIB(PARENT(to)) = (to); \
         }


/*

  PROCEDURE
  Insert - Insert a node into a tree.

  DESCRIPTION
  This routine inserts the passed node into the passed tree, using the
  default comparison node routine.

*/
static void
Insert(
  Tree *tree,	       	/* the tree into which to insert the node */
  Node *x		/* the node to insert */
)
{
  Node *y, *z;

  /* first insert the new node at the correct position */

  /* starting point for search */
  z = ROOT(tree);

  /* backtrack copy */
  y = NIL(tree);

  /* visit nodes until run off of tree */
  while (z != NIL(tree))
  {
    y = z;
    if ( tree->cmp(x->data, z->data) < 0 ) z = LEFT_SIB(z);
    else                           z = RIGHT_SIB(z);
  }

  /* ran into a leaf.  y points to leaf */
  /* attach x to p */
  PARENT(x) = y;

  /* fix child pointers for y */
  if (y == NIL(tree))
  {
    ROOT(tree) = x;
  }
  else
  {
    if ( tree->cmp(x->data, y->data) < 0 ) LEFT_SIB(y)  = x;
    else                              RIGHT_SIB(y) = x;
  }

  COLOR(x) = RED;

  /* repair tree so that it's legal */
  while (x != ROOT(tree) && COLOR(PARENT(x)) == RED)
  {
    if (PARENT(x) == LEFT_SIB(PARENT(PARENT(x))))
    {
      y = RIGHT_SIB(PARENT(PARENT(x)));
      if (COLOR(y) == RED)
      {
	COLOR(PARENT(x)) = BLACK;
	COLOR(y) = BLACK;
	COLOR(PARENT(PARENT(x))) = RED;
	x = PARENT(PARENT(x));
      }
      else
      {
	if (x == RIGHT_SIB(PARENT(x)))
	{
	  x = PARENT(x);
	  LeftRotate(tree, x);
	}
	COLOR(PARENT(x)) = BLACK;
	COLOR(PARENT(PARENT(x))) = RED;
	RightRotate(tree, PARENT(PARENT(x)));
      }
    }
    else
    {
      y = LEFT_SIB(PARENT(PARENT(x)));
      if (COLOR(y) == RED)
      {
	COLOR(PARENT(x)) = BLACK;
	COLOR(y) = BLACK;
	COLOR(PARENT(PARENT(x))) = RED;
	x = PARENT(PARENT(x));
      }
      else
      {
	if (x == LEFT_SIB(PARENT(x)))
	{
	  x = PARENT(x);
	  RightRotate(tree, x);
	}
	COLOR(PARENT(x)) = BLACK;
	COLOR(PARENT(PARENT(x))) = RED;
	LeftRotate(tree, PARENT(PARENT(x)));
      }
    }
  }

  COLOR(ROOT(tree)) = BLACK;
  COUNT(tree)++;
}

/*

  PROCEDURE
  Delete - Remove a node from a tree.

  DESCRIPTION
  @code{Delete} removes a node from a tree.  The node removed may not
  be that requested to be removed, for algorithmic efficiency's sake.
  in this case, the data is swapped with the node that is removed, so
  everything works as expected.  It does not deallocate the memory
  associated with the node.

  RETURNS
  It returns a pointer to the node that was actually removed.

*/
static Node *
Delete(
  Tree *tree,		/* the tree from which to delete the node */
  Node *z		/* the node to delete */
)
{
  Node *x, *y;

  /*
     determine which node to splice out. if z has only one descendent,
     splice it out.  if it has two, its successor is guaranteed to
     only have one; that's spliced out and swapped with z.
  */

  if (LEFT_SIB(z) == NIL(tree) || RIGHT_SIB(z) == NIL(tree))
    y = z;
  else
    y = Successor(tree, z);

  /* node to splice out has only one descendant.  grab it */
  x = (LEFT_SIB(y) != NIL(tree)) ? LEFT_SIB(y) : RIGHT_SIB(y);
  RedirectPARENT(tree, x, y);

  /* if y wasn't the node to be deleted, it's swapped with the one that is */
  if (y != z)
  {
    DATA(z) = DATA(y);
  }

  if (COLOR(y) == BLACK)
  {
    while (x != ROOT(tree) && COLOR(x) == BLACK)
    {
      if (x == LEFT_SIB(PARENT(x)))
      {
        Node *w = RIGHT_SIB(PARENT(x));
        if (COLOR(w) == RED)
        {
          COLOR(w) = BLACK;
          COLOR(PARENT(x)) = RED;
          LeftRotate(tree, PARENT(x));
          w = RIGHT_SIB(PARENT(x));
        }
        if (COLOR(LEFT_SIB(w)) == BLACK && COLOR(RIGHT_SIB(w)) == BLACK)
        {
          COLOR(w) = RED;
          x = PARENT(x);
        }
        else
        {
          if (COLOR(RIGHT_SIB(w)) == BLACK)
          {
            COLOR(LEFT_SIB(w)) = BLACK;
            COLOR(w) = RED;
            RightRotate(tree, w);
            w = RIGHT_SIB(PARENT(x));
          }
          COLOR(w) = COLOR(PARENT(x));
          COLOR(PARENT(x)) = BLACK;
          COLOR(RIGHT_SIB(w)) = BLACK;
          LeftRotate(tree, PARENT(x));
          x = ROOT(tree);
        }
      }
      else
      {
        Node *w = LEFT_SIB(PARENT(x));
        if (COLOR(w) == RED)
        {
          COLOR(w) = BLACK;
          COLOR(PARENT(x)) = RED;
          RightRotate(tree, PARENT(x));
          w = LEFT_SIB(PARENT(x));
        }
        if (COLOR(RIGHT_SIB(w)) == BLACK && COLOR(LEFT_SIB(w)) == BLACK)
        {
          COLOR(w) = RED;
          x = PARENT(x);
        }
        else
        {
          if (COLOR(LEFT_SIB(w)) == BLACK)
          {
            COLOR(RIGHT_SIB(w)) = BLACK;
            COLOR(w) = RED;
            LeftRotate(tree, w);
            w = LEFT_SIB(PARENT(x));
          }
          COLOR(w) = COLOR(PARENT(x));
          COLOR(PARENT(x)) = BLACK;
          COLOR(LEFT_SIB(w)) = BLACK;
          RightRotate(tree, PARENT(x));
          x = ROOT(tree);
        }
      }
    }

    COLOR(x) = BLACK;
  }
  COUNT(tree)--;
  return y;
}

/*

  PROCEDURE
    LeftRotate - Perform a left rotation of a tree about a node.

*/
static void
LeftRotate(
  Tree *tree,		/* the tree to rotate */
  Node *x		/* the pivot node */
)
{
  Node *y;

  /* get right node */
  y = RIGHT_SIB(x);

  /* move y's left node to x's right */
  RIGHT_SIB(x) = LEFT_SIB(y);

  /* direct x's new right node's parent link to x */
  if (RIGHT_SIB(x) != NIL(tree)) PARENT(RIGHT_SIB(x)) = x;

  /* y now has x's old parent as its own */
  RedirectPARENT(tree, y, x);

  /* establish new relationship between x and y */
  LEFT_SIB(y) = x;
  PARENT(x) = y;
}


/*

  PROCEDURE
  RightRotate - Perform a right rotation of a tree about a node.

*/
static void
RightRotate(
  Tree *tree,		/* the tree to rotate */
  Node *y		/* the pivot node */
)
{
  Node *x;

  /* get left node */
  x = LEFT_SIB(y);

  /* move x's right node over to y's left */
  LEFT_SIB(y) = RIGHT_SIB(x);

  /* direct y's new left node's parent link to y */
  if (LEFT_SIB(y) != NIL(tree)) PARENT(LEFT_SIB(y)) = y;

  /* x now has y's old parent as its own */
  RedirectPARENT(tree, x, y);

  /* establish new relationship between x and y */
  RIGHT_SIB(x) = y;
  PARENT(y) = x;
}


/*

  PROCEDURE
  Successor - Find the next in-order node to a node.

*/
static Node *
Successor(
  Tree *tree,		/* the tree to traverse */
  Node *x		/* the origin node */
)
{
  Node *y;

  if (RIGHT_SIB(x) != NIL(tree)) return Minimum(tree, RIGHT_SIB(x));
  y = PARENT(x);
  while (y != NIL(tree) && x == RIGHT_SIB(y))
  {
    x = y;
    y = PARENT(y);
  }

  return y;
}

/*

  PROCEDURE
  Predecessor - Find the previous in-order node to a node.

*/
static Node *
Predecessor(
  Tree *tree,		/* the tree to traverse */
  Node *x		/* the origin node */
)
{
  Node *y;

  if (LEFT_SIB(x) != NIL(tree)) return Maximum(tree, LEFT_SIB(x));
  y = PARENT(x);
  while (y != NIL(tree) && x == LEFT_SIB(y))
  {
    x = y;
    y = PARENT(y);
  }

  return y;
}

/*

  PROCEDURE
  Minimum - Return the left most node in the tree.

*/
static Node *
Minimum(
  Tree *tree,		/* the tree to traverse */
  Node *x		/* the origin node */
)
{
  while (LEFT_SIB(x) != NIL(tree)) x = LEFT_SIB(x);
  return x;
}

/*

  PROCEDURE
  Maximum - Return the right most node in the tree.

*/
static Node *
Maximum(
  Tree *tree,		/* the tree to traverse */
  Node *x		/* the origin node */
)
{
  while (RIGHT_SIB(x) != NIL(tree)) x = RIGHT_SIB(x);
  return x;
}


/*

  PROCEDURE
  newNode - Allocate a new node and copy data into it.

*/
static Node *
newNode(
  Tree *tree,		/* the tree that will contain the node */
  void *data		/* the data that the new node will contain */
)
{
  Node *node = (Node *) malloc(sizeof(Node));

  if (!node)
    return NULL;

  DATA(node) = data;
  LEFT_SIB(node) = RIGHT_SIB(node) = NIL(tree);
  return node;
}

/*

  PROCEDURE
  deleteNode - Delete a node.

  DESCRIPTION
  This routine deallocates memory associated with a node.  If asked to
  delete a @code{NULL} node, it just returns.

*/
static void
deleteNode(
  Node *node		/* the node to delete */
)
{
  if (node) free(node);
}


/*

  PROCEDURE
  Search - Search a tree for a node with comparable data.

  DESCRIPTION
  This routine will search a tree for the node which has data
  comparable to that passed, using the passed comparison routine.  The
  start node need not be the root of the tree, allowing for subtree
  searches.

  RETURNS
  It returns the node which matches, else @code{NIL(tree)}.

*/
static Node *
Search(
  Tree *tree,		/* the tree to search */
  Node *x,		/* the node at which to start */
  const void *data,	/* the data to search for */
  int (*cmp)(const void *, const void *)
                        /* the address of a node comparison routine */
)
{
  int c;

  while ( x != NIL(tree) && (c = cmp(data, DATA(x))) )
     x = c < 0 ? LEFT_SIB(x) : RIGHT_SIB(x);

  return x;
}

/*

  PROCEDURE
  BndSearch - Search a (sub)tree for a node with comporable data,
  keeping track of siblings.

  DESCRIPTION
  This routine will search a tree for the node which has data
  comparable to that which is passed, using the passed comparison
  routine.  If the node does not exist, it returns the predecessor and
  successor nodes.  the start node need not be the root of the tree,
  allowing for sub-tree searches.

  RETURNS
  It returns a pointer to the found node upon sucess.  If no node is
  found, it sets the parameters next and/or prev to point to the
  successor and predecessor nodes.  If those nodes don't exist, the
  pointers are set to @code{NIL(tree)}.

*/
static Node *
BndSearch(
  Tree *tree,		/* the tree to search */
  Node *x,		/* the node at which to start the search */
  Node **prev,		/* where to stick a pointer to the predecessor node */
  Node **next,		/* where to stick a pointer to the successor node */
  const void *data,	/* the data to search for */
  int (*cmp)(const void *, const void *)
                        /* the address of a node comparison function */
)
{
  int c;
  *prev = *next = NIL(tree);

  while ( x != NIL(tree) && (c = cmp(data, DATA(x))) )
  {
    if (c < 0)
    {
      *next = x;
      x = LEFT_SIB(x);
    }
    else
    {
      *prev = x;
      x = RIGHT_SIB(x);
    }
  }

  return x;
}

/*

  PROCEDURE
  Traverse - Walk a tree, processing each node.

  DESCRIPTION
  @code{Traverse} quickly walks a tree in a prescribed order, performing a
  user supplied action on each node, in-order.  If the action function
  returns non-zero, the walk is aborted.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
static int
Traverse(
  Tree *tree,			/* the tree to walk */
  Node *node,			/* the node to start at, must be
				   @code{ROOT(tree)} */
  int (*action)(void *),	/* the action to perform at each node */
  SiblingOrder sibling_order	/* the order in which to traverse the tree */
)
{
  int val = 0;

  if (! action || node == NIL(tree)) return 0;

  /* find extremum in opposite direction of travel */
  if (sibling_order == LEFT_TO_RIGHT)
  {
    node = Minimum(tree, node);
    while ( node != NIL(tree) && ! (val = action(DATA(node))) )
      node = Successor(tree, node);
  }
  else
  {
    node = Maximum(tree, node);
    while ( node != NIL(tree) && ! (val = action(DATA(node))) )
      node = Predecessor(tree, node);
  }

  return val;
}

/*

  PROCEDURE
  uTraverse - Walk a tree, processing each node.

  DESCRIPTION
  @code{uTraverse} quickly walks a tree in a prescribed order,
  performing a user supplied action on each node, in-order.  If the
  action function returns non-zero, the walk is aborted.  This routine
  differs from Traverse in that it can pass along a pointer provided
  by the calling routine to the action routine, allowing arbitrary
  data to be available to the action routine.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
static int
uTraverse(
  Tree *tree,			/* the tree to walk */
  Node *node,			/* the node to start at, must be
				   @code{ROOT(tree)} */
  int (*action)(void *, void *udata ),	/* the action to perform at
					   each node */
  void *udata,			/* a pointer to data to be passed to the action
				   routine */
  SiblingOrder sibling_order	/* the order in which to traverse the tree */
)
{
  int val = 0;

  if (! action || node == NIL(tree)) return 0;

  /* find extremum in opposite direction of travel */
  if (sibling_order == LEFT_TO_RIGHT)
  {
    node = Minimum(tree, node);
    while ( node != NIL(tree) && ! (val = action( DATA(node), udata )) )
      node = Successor(tree, node);
  }
  else
  {
    node = Maximum(tree, node);
    while ( node != NIL(tree) && ! (val = action( DATA(node), udata )) )
      node = Predecessor(tree, node);
  }

  return val;
}

/*

  PROCEDURE
  Walk - Walk a tree, performing an action at specified traversals of
  each node.

  DESCRIPTION
  @code{Walk} traverse a tree, calling a user supplied action routine
  at selectable traversals of each node (any combination of pre-order,
  in-order, or post-order).  The action routine is informed of which
  node traversal and which tree level it is called from. the walk is
  aborted if the action routine returns non-zero.

  The traversals at which the action routine is called at are
  specified by the logical or of @code{PRE_ORDER}, @code{IN_ORDER}, or
  @code{POST_ORDER}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
static int
Walk(
  Tree *tree,		/* the tree to walk */
  Node *node,		/* the node at which to start.  must be @code{ROOT(tree)} */

  int (*action)(void *data, Visit visit, unsigned long level),  /* the action routine */

  SiblingOrder sibling_order,	/* the order in which to walk */
  Visit visit,			/* the traversal(s) at which to call the action
				   routine */
  unsigned long level		/* the level of the passed node.  @samp{0} if the
				   root node. */
)
{
  int val;
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  if ( (node != NIL(tree)) && action )
  {
    if (visit & PRE_ORDER && (val = action(DATA(node), PRE_ORDER, level)) )
      return val;

    if ( ( val = Walk(tree, DESCEND(node,dir1), action, sibling_order,
         visit, level+1)) )
      return val;

    if (visit & IN_ORDER && (val = action(DATA(node), IN_ORDER, level)) )
      return val;

    if ( (val = Walk(tree, DESCEND(node,dir2), action, sibling_order,
       visit, level+1)) )
      return val;

    if (visit & POST_ORDER && (val = action(DATA(node), POST_ORDER, level)) )
      return val;
  }

  return 0;
}

/*

  PROCEDURE
  uWalk - Walk a tree, performing an action at specified traversals of
  each node.

  DESCRIPTION
  @code{uWalk} traverse a tree, calling a user supplied action routine
  at selectable traversals of each node (any combination of pre-order,
  in-order, or post-order).  The action routine is informed of which
  node traversal and which tree level it is called from. the walk is
  aborted if the action routine returns non-zero.  This routine
  differs from @code{Walk} in that it can pass along a pointer
  provided by the calling routine to the action routine, allowing
  arbitrary data to be available to the action routine.

  The traversals at which the action routine is called at are
  specified by the logical or of @code{PRE_ORDER}, @code{IN_ORDER}, or
  @code{POST_ORDER}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
static int
uWalk(
  Tree *tree,			/* the tree to walk */
  Node *node,			/* the node at which to start. must be
				   @code{ROOT(tree)} */

  int (*action)(void *data, Visit visit, unsigned long level, void *udata),	/* the action routine */

  void *udata,			/* a pointer to data to be passed to the action
				   routine */

  SiblingOrder sibling_order,	/* the order in which to walk */

  Visit visit,			/* the traversal(s) at which to call the action
				   routine */
  unsigned long level		/* the level of the passed node.
				   @samp{0} if the root node. */
)
{
  int val;
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  if ( (node != NIL(tree)) && action )
  {
    if ( visit & PRE_ORDER &&
	( val = action(DATA(node), PRE_ORDER, level, udata) ) )
      return val;

    if ( (val = uWalk(tree, DESCEND(node,dir1), action, udata, sibling_order,
         visit, level+1)) )
      return val;

    if ( visit & IN_ORDER
	&& ( val = action(DATA(node), IN_ORDER, level, udata)) )
      return val;

    if ( ( val = uWalk(tree, DESCEND(node,dir2), action, udata, sibling_order,
       visit, level+1)) )
      return val;

    if (visit & POST_ORDER &&
	( val = action(DATA(node), POST_ORDER, level, udata) ) )
      return val;
  }

  return 0;
}

/*

  PROCEDURE
  Free_post_order - Free a tree, in post-order sequence.

  DESCRIPTION
  This routine traverses the given tree, freeing (post-order) as we
  go.  It's the fastest deletion routine.  Use it when it doesn't
  matter in what order the data are deleted.

  The node data is passed to a user-supplied free routine, if available.

*/
static void
Free_post_order(
  Tree *tree,			/* the tree to destroy */
  void (*nfree)(void *),	/* a routine to free the node data,
				   may be @code{RBTREE_NULL_DELETE} */
  SiblingOrder sibling_order	/* the order in which to free the tree */
)
{
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  Node *node = ROOT(tree);

  while (node != NIL(tree))
  {
    if (DESCEND(node,dir1) != NIL(tree))
      node = DESCEND(node,dir1);
    else  if (DESCEND(node,dir2) != NIL(tree))
      node = DESCEND(node,dir2);
    else
    {
      Node *parent = PARENT(node);

      if (nfree) nfree(DATA(node));
      if (node == LEFT_SIB(parent)) LEFT_SIB(parent)  = NIL(tree);
      else                      RIGHT_SIB(parent) = NIL(tree);
      deleteNode(node);
      node = parent;
    }
  }
}


/*

  PROCEDURE
  Free_in_order - Traverse a tree, freeing nodes in-order.

  DESCRIPTION
  Traverse the given tree, freeing (in-order) as we go.  Recursive, so
  it uses more resources than @code{Free_post_order}.

*/
static void
Free_in_order(
  Tree *tree,			/* the tree to free */
  Node *node,			/* the starting node, must be @code{ROOT(tree)} */
  void (*nfree)(void *),	/* a free routine for node data, may
				   be @code{RBTREE_NULL_DELETE} */
  SiblingOrder sibling_order	/* the order in which to traverse the tree */
)
{
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  if (node != NIL(tree))
  {
    Free_in_order(tree, DESCEND(node,dir1), nfree, sibling_order);

    if (nfree) nfree(DATA(node));

    Free_in_order(tree, DESCEND(node,dir2), nfree, sibling_order);

    deleteNode(node);
  }
}


/*

  PROCEDURE
  uFree_post_order - Free a tree, in post-order sequence.

  DESCRIPTION
  This routine traverses the given tree, freeing (post-order) as we
  go.  It's the fastest deletion routine.  Use it when it doesn't
  matter in what order the data are deleted.  This routine differs
  from @code{Free_post_order} in that it can pass along a pointer
  provided by the calling routine to the action routine, allowing
  arbitrary data to be available to the action routine.  The node data
  is passed to a user-supplied free routine, if available.

*/
static void
uFree_post_order(
  Tree *tree,			/* the tree to destroy */
  void (*nfree)(void *, void *),	/* a routine to free the node data,
				   may be @code{RBTREE_NULL_DELETE} */
  void *udata,			/* a pointer to data to be
				   passed to the action routine */
  SiblingOrder sibling_order	/* the order in which to free the tree */
)
{
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  Node *node = ROOT(tree);

  while (node != NIL(tree))
  {
    if (DESCEND(node,dir1) != NIL(tree))
      node = DESCEND(node,dir1);
    else  if (DESCEND(node,dir2) != NIL(tree))
      node = DESCEND(node,dir2);
    else
    {
      Node *parent = PARENT(node);

      if (nfree) nfree(DATA(node), udata);
      if (node == LEFT_SIB(parent)) LEFT_SIB(parent)  = NIL(tree);
      else                      RIGHT_SIB(parent) = NIL(tree);
      deleteNode(node);
      node = parent;
    }
  }
}


/*

  PROCEDURE
  uFree_in_order - Traverse a tree, freeing nodes in-order.

  DESCRIPTION
  Traverse the given tree, freeing (in-order) as we go.  Recursive, so
  it uses more resources than @code{Free_post_order}.
  This routine
  differs from @code{Free_in_order} in that it can pass along a pointer
  provided by the calling routine to the action routine, allowing
  arbitrary data to be available to the action routine.
*/
static void
uFree_in_order(
  Tree *tree,			/* the tree to free */
  Node *node,			/* the starting node, must be @code{ROOT(tree)} */
  void (*nfree)(void *, void *),	/* a free routine for node data, may
				   be @code{RBTREE_NULL_DELETE} */
  void *udata,			/* a pointer to data to be
				   passed to the action routine */
  SiblingOrder sibling_order	/* the order in which to traverse the tree */
)
{
  Direction dir1 = sibling_order == LEFT_TO_RIGHT ? LEFT : RIGHT;
  Direction dir2 = dir1 == LEFT ? RIGHT : LEFT;

  if (node != NIL(tree))
  {
    uFree_in_order(tree, DESCEND(node,dir1), nfree, udata, sibling_order);

    if (nfree) nfree( DATA(node), udata );

    uFree_in_order(tree, DESCEND(node,dir2), nfree, udata, sibling_order);

    deleteNode(node);
  }
}


/*

  PROCEDURE
  JoinTrees - Join two trees.

  DESCRIPTION
  This routine joins two trees, detaching nodes from the second and
  inserting them in the first.  As the second tree is completely
  gutted, the routine doesn't spend time keeping its redblack nature
  intact during the process.

*/
static void
JoinTrees(
  Tree *tree1,		/* the tree to merge the second into */
  Tree *tree2		/* the tree to merge into the first tree */
)
{
  Node *node = ROOT(tree2);

  while (node != NIL(tree2))
  {
    if      (LEFT_SIB(node)  != NIL(tree2))
      node = LEFT_SIB(node);
    else if (RIGHT_SIB(node) != NIL(tree2))
      node = RIGHT_SIB(node);
    else
    {
      Node *parent = PARENT(node);

      if (node == LEFT_SIB(parent)) LEFT_SIB(parent)  = NIL(tree2);
      else                      RIGHT_SIB(parent) = NIL(tree2);
      Insert(tree1, node);
      node = parent;
    }
  }
  COUNT(tree2) = 0;
  ROOT(tree2) = NIL(tree2);
}
